import asyncio

from conftest import vcr_c

from aiotx.clients import AioTxBSCClient


@vcr_c.use_cassette("tests/fixtures/cassettes/bsc/test_async_monitoring.yaml")
async def test_async_monitoring(bsc_client: AioTxBSCClient):
    blocks = []
    transactions = []

    @bsc_client.monitor.on_block
    async def handle_block(block):
        blocks.append(block)

    @bsc_client.monitor.on_transaction
    async def handle_transaction(transaction):
        transactions.append(transaction)

    await bsc_client.start_monitoring(584)
    try:
        await asyncio.sleep(3)
    except KeyboardInterrupt:
        bsc_client.stop_monitoring()

    assert len(blocks) > 0
    assert len(transactions) > 0

    assert 584 in blocks
    assert 585 in blocks
    assert "0x2a9d5aea2c8e4759e4e664f48e0dac79e76cb024e670251ddb2788144cee6f43" in [
        tx["hash"] for tx in transactions
    ]
    assert "0xa45ccbf3c820f7bec2809461a8b7f1856246ed3b650ec2840bfbc303f118b76c" in [
        tx["hash"] for tx in transactions
    ]

    for tx in transactions:
        assert "aiotx_decoded_input" in tx.keys()



from ._keystore import generate_keystore_key, generate_new_keystore
from ._mnemonic import mnemonic_is_valid, mnemonic_new, mnemonic_to_wallet_key
from ._utils import private_key_to_public_key, verify_sign

__all__ = [
    "mnemonic_new",
    "mnemonic_to_wallet_key",
    "mnemonic_is_valid",
    "generate_new_keystore",
    "generate_keystore_key",
    "private_key_to_public_key",
    "verify_sign",
    "generate_key_pair",
]



from sqlalchemy import BigInteger, Boolean, Column, Integer, String
from sqlalchemy.orm import declarative_base

Base = declarative_base()


def create_address_model(currency_name):
    class Address(Base):
        __tablename__ = f"{currency_name}_addresses"
        __table_args__ = {"extend_existing": True}

        address = Column(String(255), primary_key=True)
        block_number = Column(Integer)

    return Address


def create_utxo_model(currency_name):
    class UTXO(Base):
        __tablename__ = f"{currency_name}_utxo"
        __table_args__ = {"extend_existing": True}

        tx_id = Column(String(255), primary_key=True)
        output_n = Column(Integer, primary_key=True)
        address = Column(String(255))
        amount_satoshi = Column(BigInteger)
        used = Column(Boolean, default=False)

    return UTXO


def create_last_block_model(currency_name):
    class LastBlock(Base):
        __tablename__ = f"{currency_name}_last_block"
        __table_args__ = {"extend_existing": True}

        block_number = Column(Integer, primary_key=True)

    return LastBlock



import os
from hashlib import pbkdf2_hmac
from typing import Tuple

from nacl.bindings import crypto_box_seed_keypair


def generate_keystore_key(password: str, salt: bytes) -> Tuple[bytes, bytes]:
    """
    :rtype: (bytes(public_key), bytes(secret_key))
    """
    secret = pbkdf2_hmac("sha512", password.encode("utf-8"), salt, 400_000, 32)
    return crypto_box_seed_keypair(secret)


def generate_new_keystore(password: str, version: int = 1):
    salt = os.urandom(32)
    pub_k, _ = generate_keystore_key(password, salt)

    return {"version": version, "salt": salt.hex(), "public_key": pub_k.hex()}



from abc import ABC

from ..boc import Cell
from ..utils import Address


class Contract(ABC):
    def __init__(self, **kwargs):
        self.options = kwargs
        self._address = Address(kwargs["address"]) if "address" in kwargs else None
        if "wc" not in kwargs:
            kwargs["wc"] = self._address.wc if self._address is not None else 0

    @property
    def address(self):
        if self._address is None:
            self._address = self.create_state_init()["address"]

        return self._address

    def create_state_init(self):
        code_cell = self.create_code_cell()
        data_cell = self.create_data_cell()
        state_init = self.__create_state_init(code_cell, data_cell)
        state_init_hash = state_init.bytes_hash()

        address = Address(str(self.options["wc"]) + ":" + state_init_hash.hex())

        return {
            "code": code_cell,
            "data": data_cell,
            "address": address,
            "state_init": state_init,
        }

    def create_code_cell(self):
        if "code" not in self.options or self.options["code"] is None:
            raise Exception("Contract: options.code is not defined")
        return self.options["code"]

    def create_data_cell(self):
        return Cell()

    def create_init_external_message(self):
        create_state_init = self.create_state_init()
        state_init = create_state_init["state_init"]
        address = create_state_init["address"]
        code = create_state_init["code"]
        data = create_state_init["data"]
        header = Contract.create_external_message_header(address)
        external_message = Contract.create_common_msg_info(header, state_init)
        return {
            "address": address,
            "message": external_message,
            "state_init": state_init,
            "code": code,
            "data": data,
        }

    @classmethod
    def create_external_message_header(cls, dest, src=None, import_fee=0):
        message = Cell()
        message.bits.write_uint(2, 2)
        message.bits.write_address(Address(src) if src else None)
        message.bits.write_address(Address(dest))
        message.bits.write_grams(import_fee)
        return message

    @classmethod
    def create_internal_message_header(
        cls,
        dest,
        grams=0,
        ihr_disabled=True,
        bounce=None,
        bounced=False,
        src=None,
        currency_collection=None,
        ihr_fees=0,
        fwd_fees=0,
        created_lt=0,
        created_at=0,
    ):
        message = Cell()
        message.bits.write_bit(0)
        message.bits.write_bit(ihr_disabled)

        if bounce is not None:
            message.bits.write_bit(bounce)
        else:
            message.bits.write_bit(Address(dest).is_bounceable)
        message.bits.write_bit(bounced)
        message.bits.write_address(Address(src) if src else None)
        message.bits.write_address(Address(dest))
        message.bits.write_grams(grams)
        if currency_collection:
            raise Exception("Currency collections are not implemented yet")

        message.bits.write_bit(bool(currency_collection))
        message.bits.write_grams(ihr_fees)
        message.bits.write_grams(fwd_fees)
        message.bits.write_uint(created_lt, 64)
        message.bits.write_uint(created_at, 32)
        return message

    @classmethod
    def create_common_msg_info(cls, header, state_init=None, body=None):
        common_msg_info = Cell()
        common_msg_info.write_cell(header)
        if state_init:
            common_msg_info.bits.write_bit(1)
            if (
                common_msg_info.bits.get_free_bits() - 1
                >= state_init.bits.get_used_bits()
            ):
                common_msg_info.bits.write_bit(0)
                common_msg_info.write_cell(state_init)
            else:
                common_msg_info.bits.write_bit(1)
                common_msg_info.refs.append(state_init)
        else:
            common_msg_info.bits.write_bit(0)

        if body:
            if common_msg_info.bits.get_free_bits() >= body.bits.get_used_bits():
                common_msg_info.bits.write_bit(0)
                common_msg_info.write_cell(body)
            else:
                common_msg_info.bits.write_bit(1)
                common_msg_info.refs.append(body)
        else:
            common_msg_info.bits.write_bit(0)

        return common_msg_info

    def __create_state_init(
        self, code, data, library=None, split_depth=None, ticktock=None
    ):
        if library or split_depth or ticktock:
            raise Exception(
                "Library/SplitDepth/Ticktock in state init is not implemented"
            )

        state_init = Cell()
        settings = bytes(
            "".join(
                [
                    "1" if i else "0"
                    for i in [
                        bool(split_depth),
                        bool(ticktock),
                        bool(code),
                        bool(data),
                        bool(library),
                    ]
                ]
            ),
            "utf-8",
        )
        state_init.bits.write_bit_array(settings)

        if code:
            state_init.refs.append(code)
        if data:
            state_init.refs.append(data)
        if library:
            state_init.refs.append(library)
        return state_init



import asyncio

from aiotx.clients import AioTxTONClient
from aiotx.utils.tonsdk.contract.wallet import WalletVersionEnum

# For bulk send we should use HighloadWalletV2Contract and deploy it before the first send!

ton_client = AioTxTONClient(
    "https://testnet.toncenter.com/api/v2",
    workchain=0,
    wallet_version=WalletVersionEnum.hv2,
)


recipients_list = [
    {
        "address": "0QAEhA1CupMp7uMOUfHHoh7sqAMNu1xQOydf8fQf-ATpkbpT",
        "amount": 1,
        "payload": "Hello, recipient 1!",
        "send_mode": 1,
    },
    {
        "address": "UQDlTHD4T79EyT96gkYNKd3iuRd2__6gGh2PCKpU57jSWQ7j",
        "amount": 10,
        "payload": "Hello, recipient 5!",
        "send_mode": 1,
    },
]

# You should deploy the wallet before you will be able to start using it, wallet should have balance for that to deploy itself!

tx_id = asyncio.run(
    ton_client.deploy_wallet(
        "ancient reason board glue post sea write shrimp feel motion win away pizza sword mobile ethics film lawsuit nephew cloud crash okay hero step"
    )
)

tx_id = asyncio.run(
    ton_client.send_bulk(
        "ancient reason board glue post sea write shrimp feel motion win away pizza sword mobile ethics film lawsuit nephew cloud crash okay hero step",
        recipients_list,
    )
)

print("tx_id", tx_id)



import asyncio

from aiogram import Bot, Dispatcher

from aiotx.clients import AioTxBSCClient

# Create an instance of the aiogram bot
bot = Bot(token="BOT TOKEN")
dispatcher = Dispatcher()

# Create an instance of the AioTxBSCClient
bsc_client = AioTxBSCClient("https://bsc-testnet-rpc.publicnode.com")


# Define the block handler
@bsc_client.monitor.on_block
async def handle_block(block):
    block_number = block
    chat_id = "5454053704"  # Replace with the actual chat ID
    message = f"New block: {block_number}"
    await bot.send_message(chat_id=chat_id, text=message)


# Define the transaction handler (optional)
@bsc_client.monitor.on_transaction
async def handle_transaction(transaction):
    transaction_hash = transaction["hash"]
    chat_id = "5454053704"  # Replace with the actual chat ID
    message = f"New transaction: {transaction_hash}"
    await bot.send_message(chat_id=chat_id, text=message)


async def main():
    monitoring_task = asyncio.create_task(bsc_client.start_monitoring(584))
    await asyncio.gather(monitoring_task, dispatcher.start_polling(bot))


if __name__ == "__main__":
    asyncio.run(main())



import os
import sys

import pytest
from conftest import vcr_c

from aiotx.clients import AioTxBTCClient
from aiotx.exceptions import NotImplementedError, RpcConnectionError
from aiotx.types import FeeEstimate

TEST_BTC_WALLET_PRIVATE_KEY = os.getenv("TEST_BTC_WALLET_PRIVATE_KEY")
assert TEST_BTC_WALLET_PRIVATE_KEY is not None, "add TEST_BTC_WALLET_PRIVATE_KEY"
TEST_BTC_ADDRESS = "tb1qswslzcdulvlk62gdrg8wa0sw36f938h2pyfreh"


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("btc/send_transaction.yaml")
async def test_send_transaction(btc_client: AioTxBTCClient):
    await btc_client.monitor._add_new_address(TEST_BTC_ADDRESS)
    await btc_client.monitor._add_new_utxo(
        TEST_BTC_ADDRESS,
        "25d56f693d5c4d00d3f98e58c8bd66e8db930e38c8a556bd67737b66cbf31ab9",
        16263,
        0,
    )

    amount = 1000
    fee = btc_client.to_satoshi(0.00001)
    tx_id = await btc_client.send(
        TEST_BTC_WALLET_PRIVATE_KEY,
        "tb1q4tf08gc2vf8dgfqtd3jen3s5tddz6uu0mnyf90",
        amount,
        fee,
    )
    utxo_list = await btc_client.monitor._get_utxo_data(TEST_BTC_ADDRESS)
    assert len(utxo_list) == 1
    assert (
        utxo_list[0].tx_id
        == "4f9a64d28ae22134379f7da50282d7ca2ead8a1f8378b20e25defcaaa5c59228"
    )
    assert utxo_list[0].amount_satoshi == 14263
    assert tx_id == "4f9a64d28ae22134379f7da50282d7ca2ead8a1f8378b20e25defcaaa5c59228"
    await btc_client.monitor._delete_utxo(
        "25d56f693d5c4d00d3f98e58c8bd66e8db930e38c8a556bd67737b66cbf31ab9", 0
    )


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("btc/send_to_all_address_types.yaml")
async def test_send_to_all_address_types(btc_client: AioTxBTCClient):
    fee = btc_client.to_satoshi(0.00006)
    await btc_client.monitor._add_new_address(TEST_BTC_ADDRESS)
    await btc_client.monitor._add_new_utxo(
        TEST_BTC_ADDRESS,
        "25d56f693d5c4d00d3f98e58c8bd66e8db930e38c8a556bd67737b66cbf31ab9",
        16263,
        0,
    )
    amount = 1000
    tx_id = await btc_client.send_bulk(
        TEST_BTC_WALLET_PRIVATE_KEY,
        {
            "mtkbaiLiUH3fvGJeSzuN3kUgmJzqinLejJ": amount,  # Legacy P2PKH address
            "2NDYhjHKdPLYjD5kVZVgeo8JxDrDUcRC14Q": amount,  # P2SH address
            "tb1qq4ay9zw2mygwucz47v6hsda04a9l2yajyynh00": amount,  # Bech32 address
            "tb1pwgjc67568cu6vncgve60t6eaht0rwr4pwdymnrpezrpepg3q5t8sp5250h": amount,  # Bech32M address
        },
        fee,
    )
    utxo_list = await btc_client.monitor._get_utxo_data(TEST_BTC_ADDRESS)
    assert len(utxo_list) == 1
    assert (
        utxo_list[0].tx_id
        == "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2"
    )
    assert utxo_list[0].amount_satoshi == 6263
    assert tx_id == "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2"


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("btc/send_with_auto_fee.yaml")
async def test_send_with_auto_fee(btc_client: AioTxBTCClient):
    await btc_client.monitor._add_new_address(TEST_BTC_ADDRESS)
    await btc_client.monitor._add_new_utxo(
        TEST_BTC_ADDRESS,
        " 35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ",
        6263,
        0,
    )
    amount = 1000
    tx_id = await btc_client.send_bulk(
        TEST_BTC_WALLET_PRIVATE_KEY,
        {"mtkbaiLiUH3fvGJeSzuN3kUgmJzqinLejJ": amount},
        conf_target=1,
        estimate_mode=FeeEstimate.CONSERVATIVE,
    )

    assert tx_id == "77cb20c4b0325242b9e5f45f4850e5387dc585d6b72bb36ba65a126534436973"

    utxo_list = await btc_client.monitor._get_utxo_data(TEST_BTC_ADDRESS)
    assert len(utxo_list) == 0


@vcr_c.use_cassette("btc/test_get_balance.yaml")
async def test_get_balance(btc_client: AioTxBTCClient):
    await btc_client.monitor._add_new_address(TEST_BTC_ADDRESS)
    await btc_client.monitor._add_new_utxo(
        TEST_BTC_ADDRESS,
        "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ",
        6263,
        0,
    )

    balance = await btc_client.get_balance(TEST_BTC_ADDRESS)
    assert balance == 6263

    await btc_client.monitor._delete_utxo(
        "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ", 0
    )

    balance = await btc_client.get_balance(TEST_BTC_ADDRESS)
    assert balance == 0

    await btc_client.monitor._add_new_utxo(
        TEST_BTC_ADDRESS,
        "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ",
        6263,
        0,
    )
    balance = await btc_client.get_balance(TEST_BTC_ADDRESS)
    assert balance == 6263

    await btc_client.monitor._mark_utxo_used(
        "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ", 0
    )

    balance = await btc_client.get_balance(TEST_BTC_ADDRESS)
    assert balance == 0


@pytest.mark.mysql
@vcr_c.use_cassette("btc/test_get_balance_mysql.yaml")
async def test_get_balance_mysql(btc_client_mysql: AioTxBTCClient):
    await btc_client_mysql.monitor._add_new_address(TEST_BTC_ADDRESS)
    await btc_client_mysql.monitor._add_new_utxo(
        TEST_BTC_ADDRESS,
        "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ",
        6263,
        0,
    )

    balance = await btc_client_mysql.get_balance(TEST_BTC_ADDRESS)
    assert balance == 6263

    await btc_client_mysql.monitor._delete_utxo(
        "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ", 0
    )

    balance = await btc_client_mysql.get_balance(TEST_BTC_ADDRESS)
    assert balance == 0

    await btc_client_mysql.monitor._add_new_utxo(
        TEST_BTC_ADDRESS,
        "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ",
        6263,
        0,
    )
    balance = await btc_client_mysql.get_balance(TEST_BTC_ADDRESS)
    assert balance == 6263

    await btc_client_mysql.monitor._mark_utxo_used(
        "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ", 0
    )

    balance = await btc_client_mysql.get_balance(TEST_BTC_ADDRESS)
    assert balance == 0


@pytest.mark.parametrize(
    "tx_id, expected_exception, expected_fee",
    [
        (
            "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2",
            None,
            6000,
        ),
        (
            "77cb20c4b0325242b9e5f45f4850e5387dc585d6b72bb36ba65a126534436973",
            None,
            5263,
        ),
        (
            "25d56f693d5c4d00d3f98e58c8bd66e8db930e38c8a556bd67737b66cbf31ab9",
            None,
            6890,
        ),
        (
            "12d9d269b4468aaf663b0712e9cef2b1a86fc7e758094af19434b0de86208611",
            None,
            648000,
        ),
        (
            "07ce6824d6c7ee226c3d311d6729d06144b0cb1fd371208b7466a6283de0c2fe",
            None,
            462000,
        ),
        (
            "2ab5a6a291ece2240642369589442029a6ae4baaefb83ff6760893e542f2f8ac",
            NotImplementedError,
            0,
        ),
        (
            "55863cc61de0c6c1c87282d3d6fb03650c0fc90ed3282191c618069cbde1d5",
            RpcConnectionError,
            0,
        ),
        (
            "863cc61de0c6c1c87282d3d6fb03650c0fc90ed3282191c618069cbde1d525",
            RpcConnectionError,
            0,
        ),
    ],
)
@vcr_c.use_cassette("btc/get_tx_fee.yaml")
async def test_get_tx_fee(
    btc_client: AioTxBTCClient, tx_id, expected_exception, expected_fee
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await btc_client.get_tx_fee(tx_id)
    else:
        fee = await btc_client.get_tx_fee(tx_id)
        assert isinstance(fee, int)
        assert fee == expected_fee


@pytest.mark.parametrize(
    "tx_id, expected_exception",
    [
        (
            "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2",
            None,
        ),
        (
            "77cb20c4b0325242b9e5f45f4850e5387dc585d6b72bb36ba65a126534436973",
            None,
        ),
        (
            "2ab5a6a291ece2240642369589442029a6ae4baaefb83ff6760893e542f2f8ac",
            None,
        ),
        (
            "55863cc61de0c6c1c87282d3d6fb03650c0fc90ed3282191c618069cbde1d5",
            RpcConnectionError,
        ),
        (
            "863cc61de0c6c1c87282d3d6fb03650c0fc90ed3282191c618069cbde1d525",
            RpcConnectionError,
        ),
    ],
)
@vcr_c.use_cassette("btc/get_raw_transaction.yaml")
async def test_get_raw_transaction(
    btc_client: AioTxBTCClient, tx_id, expected_exception
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await btc_client.get_raw_transaction(tx_id)
    else:
        tx_data = await btc_client.get_raw_transaction(tx_id)
        assert isinstance(tx_data, dict)
        assert "txid" in tx_data.keys()


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("btc/speed_up_transaction.yaml")
async def test_speed_up_transaction(btc_client: AioTxBTCClient):
    amount = btc_client.to_satoshi(3.6473502)
    await btc_client.monitor._add_new_address(TEST_BTC_ADDRESS)
    await btc_client.monitor._add_new_utxo(
        TEST_BTC_ADDRESS,
        "9e6fd5e4c946bde7f96c7f5921c46588f2462c33d68315b266373bf60735a39e",
        amount,
        0,
    )
    amount = 1000

    parent_tx_id = await btc_client.send(
        TEST_BTC_WALLET_PRIVATE_KEY,
        "tb1p2tzs7ghuanrm8mmqqjgvm2fdkkl4t6yvmupmt8wwpvsgwcj2vsasvwnjvy",
        amount,
        total_fee=int(btc_client.to_satoshi(0.00154914) / 6),
    )

    assert (
        parent_tx_id
        == "83d4a70c215017ae757b504c1c9f076b7dad31544d994b862113466d12318c48"
    )

    child_tx_id = await btc_client.speed_up_transaction_by_self_child_payment(
        TEST_BTC_WALLET_PRIVATE_KEY,
        "83d4a70c215017ae757b504c1c9f076b7dad31544d994b862113466d12318c48",
        conf_target=1,
    )
    assert (
        child_tx_id
        == "30aebb6b7f1b80ffd8700fb492e1a2820b8a04e7d6c95ddb8a163c138f6690b4"
    )


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("btc/speed_up_transaction_not_enough_outputs.yaml")
async def test_speed_up_transaction_not_enough_outputs(btc_client: AioTxBTCClient):
    await btc_client.monitor._add_new_address(TEST_BTC_ADDRESS)

    await btc_client.monitor._add_new_utxo(
        TEST_BTC_ADDRESS,
        "b68f08d7e9e98c48460475eebd820dddc956f87857f573770f20aa67e70e8df9",
        btc_client.to_satoshi(0.0006635),
        1,
    )

    amount = 500
    parent_tx_id = await btc_client.send(
        TEST_BTC_WALLET_PRIVATE_KEY,
        TEST_BTC_ADDRESS,
        amount,
        total_fee=btc_client.to_satoshi(0.0006635) - amount,
    )

    assert (
        parent_tx_id
        == "371e7db7f99ca03cdcd96c69e2822ed3c384f418f7239b388c3ad9e63adfd76f"
    )

    await btc_client.monitor._add_new_utxo(
        TEST_BTC_ADDRESS,
        "b68f08d7e9e98c48460475eebd820dddc956f87857f573770f20aa67e70e8df9",
        btc_client.to_satoshi(3.65162827),
        0,
    )

    child_tx_id = await btc_client.speed_up_transaction_by_self_child_payment(
        TEST_BTC_WALLET_PRIVATE_KEY,
        "48b4c17afda93204f0875987a9281b4bfd90c29a5590046f09a5e3d910fab8f2",
        conf_target=1,
    )
    assert (
        child_tx_id
        == "9e6fd5e4c946bde7f96c7f5921c46588f2462c33d68315b266373bf60735a39e"
    )


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("btc/bulk_send.yaml")
async def test_bulk_send(btc_client: AioTxBTCClient):
    await btc_client.monitor._add_new_address(TEST_BTC_ADDRESS)

    await btc_client.monitor._add_new_utxo(
        TEST_BTC_ADDRESS,
        "a7d8e3dea2da06c30dde62f3ead03b9260372e09b77fd74aebc7d58245a4a8bc",
        btc_client.to_satoshi(3.65416086),
        0,
    )

    tx_id = await btc_client.send_bulk(
        TEST_BTC_WALLET_PRIVATE_KEY,
        destinations={
            TEST_BTC_ADDRESS: btc_client.to_satoshi(0.0006635),
            "tb1p2tzs7ghuanrm8mmqqjgvm2fdkkl4t6yvmupmt8wwpvsgwcj2vsasvwnjvy": 600,
        },
    )
    assert tx_id == "b68f08d7e9e98c48460475eebd820dddc956f87857f573770f20aa67e70e8df9"



import os

import pytest
from conftest import vcr_c  # noqa
from vcr.errors import CannotOverwriteExistingCassetteException

from aiotx.clients import AioTxTONClient
from aiotx.exceptions import (
    AioTxError,
    BlockNotFoundError,
    InvalidArgumentError,
    RpcConnectionError,
    WrongPrivateKey,
)
from aiotx.utils.tonsdk.utils._exceptions import InvalidAddressError

TON_TEST_WALLET_MEMO = os.environ.get("TON_TEST_WALLET_MEMO")
assert TON_TEST_WALLET_MEMO is not None, "Please provide TON_TEST_WALLET_MEMO"
TON_HV_TEST_WALLET_MEMO = os.environ.get("TON_HV_TEST_WALLET_MEMO")
assert TON_HV_TEST_WALLET_MEMO is not None, "Please provide TON_HV_TEST_WALLET_MEMO"


@vcr_c.use_cassette("ton/get_last_master_block.yaml")
async def test_get_last_master_block(ton_client: AioTxTONClient):
    master_block_data = await ton_client.get_last_master_block()
    assert isinstance(master_block_data, dict)
    for key in ["shard", "workchain", "seqno", "root_hash"]:
        assert key in master_block_data.keys()


@pytest.mark.parametrize(
    "seqno, expected_exception, shards_result",
    [
        (
            19563127,
            None,
            {
                "2305843009213693952",
                "6917529027641081856",
                "-6917529027641081856",
                "-2305843009213693952",
            },
        ),
        (
            19563129,
            None,
            {
                "-2305843009213693952",
                "-6917529027641081856",
                "2305843009213693952",
                "6917529027641081856",
            },
        ),
        (
            19563128,
            None,
            {
                "-2305843009213693952",
                "-6917529027641081856",
                "2305843009213693952",
                "6917529027641081856",
            },
        ),
        (
            19566127,
            None,
            {
                "-2305843009213693952",
                "-6917529027641081856",
                "2305843009213693952",
                "6917529027641081856",
            },
        ),
        (
            19563137,
            None,
            {
                "2305843009213693952",
                "6917529027641081856",
                "-6917529027641081856",
                "-2305843009213693952",
            },
        ),
        (
            19563227,
            None,
            {
                "-2305843009213693952",
                "-6917529027641081856",
                "2305843009213693952",
                "6917529027641081856",
            },
        ),
        (
            1951663127,
            BlockNotFoundError,
            {-2305843009213693952, 6917529027641081856, -2305843009213693952},
        ),
    ],
)
@vcr_c.use_cassette("ton/get_master_block_shards.yaml")
async def test_get_master_block_shards(
    ton_client: AioTxTONClient, seqno, expected_exception, shards_result
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await ton_client.get_master_block_shards(seqno=seqno)
    else:
        shards_data = await ton_client.get_master_block_shards(seqno=seqno)
        shard_numbers = set([shard["shard"] for shard in shards_data])
        assert shard_numbers.symmetric_difference(shards_result) == set()


@pytest.mark.parametrize(
    "shard, seqno, expected_exception, tx_data_result",
    [
        (
            2305843009213693952,
            21020822,
            None,
            {
                "AwDwEqG+/w64ca+Xoc4JkJl+Iwkrs79Sjs82L2s0Vtw=",
                "HrT5S8S4zMifwDVspjEKDlSsry/6Nt9IAqmRPnWHdGU=",
            },
        ),
        (
            6917529027641081856,
            21020817,
            None,
            {"hSF1DmJJgbs/o6d9ZoGxoafRBuKChD9DEFNsXVkCQGA="},
        ),
        (
            2305843009213693952,
            21020824,
            None,
            {},
        ),
        (
            2305843009213693952,
            21023867,
            None,
            {
                "UB4KXyYA5m7cAgQD0UM50AfNhyGCQ2gAsyYyJxrlurI=",
                "bW6gqpLzLK+cvbiNsd9Njlsrvz7XlJXW2EpUAEV2/dU=",
            },
        ),
        (
            -6917529027641081856,
            21021183,
            None,
            {},
        ),
        (
            78568658568,
            0,
            BlockNotFoundError,
            {},
        ),
    ],
)
@vcr_c.use_cassette("ton/get_block_transactions.yaml")
async def test_get_block_transactions(
    ton_client: AioTxTONClient, shard, seqno, expected_exception, tx_data_result
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await ton_client.get_block_transactions(
                workchain=0, seqno=seqno, shard=shard, count=2
            )
    else:
        tx_data = await ton_client.get_block_transactions(
            workchain=0, seqno=seqno, shard=shard, count=2
        )
        tx_ids = set([shard["hash"] for shard in tx_data])
        assert tx_ids.symmetric_difference(tx_data_result) == set()


@pytest.mark.parametrize(
    "address, expected_balance, expected_exception",
    [
        (
            "EQDBNfV4DQzSyzNMw6BCTSZSoUi-CzWcYNsfhKxoDqfrwFtS",
            100917944877,
            None,
        ),
        (
            "Ef8jPzrhTYloKgTCsGgEFNx7OdH-sJ98etJnwrIVSsFxH9mu",
            92845495719,
            None,
        ),
        (
            "EQC1ZeKX1LNrlQ4bwi3je3KVM1AoZ3rkeyHM5hv9pYzmIh4v",
            24984729082,
            None,
        ),
        (
            "Ef9fwskZLEuGDfYTRAtvt9k-mEdkaIskkUOsEwPw1wzXk7zR",
            72142761735,
            None,
        ),
        (
            "0:6c058bb4a37582e27dc44c7e95cae61c8ef40fbac7783a11bc54840510b1b380",
            87235995,
            None,
        ),
        (
            "0:6c058bb4a37582e1b380",
            0,
            None,
        ),
        (
            "1:6c058bb4a37582e27dc44c7e95cae61c8ef40fbac7783a11bc54840510b1b380",
            87235995,
            AioTxError,
        ),
        (
            "Ef9fwskZLEuGDfYTRAtvt9k-mEDkaIskkUOsEwPw1wzXk7zR",
            87235995,
            InvalidArgumentError,
        ),
    ],
)
@vcr_c.use_cassette("ton/get_balance.yaml")
async def test_get_balance(
    ton_client: AioTxTONClient, address, expected_balance, expected_exception
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await ton_client.get_balance(address)
    else:
        balance = await ton_client.get_balance(address)
        assert expected_balance == balance


def get_jetton_balance_generate_cassette_name(address, jetton_master_address):
    return (
        f"ton/get_jetton_wallet_balance_{address[:8]}_{jetton_master_address[:8]}.yaml"
    )


@pytest.mark.parametrize(
    "address, jetton_master_address, expected_balance, expected_exception",
    [
        (
            "0QDlTHD4T79EyT96gkYNKd3iuRd2__6gGh2PCKpU57jSWbVp",
            "kQAiboDEv_qRrcEdrYdwbVLNOXBHwShFbtKGbQVJ2OKxY_Di",
            2111000000000,
            None,
        ),
        (
            "0QAEhA1CupMp7uMOUfHHoh7sqAMNu1xQOydf8fQf-ATpkbpT",
            "kQAiboDEv_qRrcEdrYdwbVLNOXBHwShFbtKGbQVJ2OKxY_Di",
            999999999999999997889000000000,
            None,
        ),
        (
            "EQCYRLyN3G4jePSCOnVVuutLk2pdTCjjSSkGtgOcgZ4GZjHb",
            "kQCKt2WPGX-fh0cIAz38Ljd_OKQjoZE_cqk7QrYGsNP6wfP0",
            58076874355,
            None,
        ),
        (
            "0QDBorbUtHys99DsbZ4rhfhvE7ddrC1LqUbTmRGNGVEgvVFS",
            "kQCKt2WPGX-fh0cIAz38Ljd_OKQjoZE_cqk7QrYGsNP6wfP0",
            1731564544000,
            None,
        ),
        (
            "0QDGi5-PodzUMc-WzvN7Fv1ysVUpcrAX1xKW9Feb1Y2VD4gl",
            "kQCKt2WPGX-fh0cIAz38Ljd_OKQjoZE_cqk7QrYGsNP6wfP0",
            19192552733,
            None,
        ),
        (
            "0QDGi5-PodzUMc-WzvN7Fv1ysVUpc1xKW9Feb1Y2VD4gl",
            "kQCKt2WPGX-fh0cIAz38Ljd_OKQjoZE_cqk7QrYGsNP6wfP0",
            0,
            InvalidAddressError,
        ),
        (
            "0QDGi5-PodzUMc-WzvN7Fv1ysVUpcrAX1xKW9Feb1Y2VD4gl",
            "QCKt2WPGX-fh0cIAz38Ljd_OKQjoZE_cqk7QrYGsNP6wfP0",
            0,
            RpcConnectionError,
        ),
    ],
)
async def test_get_jetton_wallet_balance(
    ton_client: AioTxTONClient,
    address,
    jetton_master_address,
    expected_balance,
    expected_exception,
):
    cassette_name = get_jetton_balance_generate_cassette_name(
        address, jetton_master_address
    )
    with vcr_c.use_cassette(cassette_name):
        if expected_exception:
            with pytest.raises(expected_exception):
                await ton_client.get_jetton_wallet_balance(
                    address, jetton_master_address
                )
        else:
            balance = await ton_client.get_jetton_wallet_balance(
                address, jetton_master_address
            )
            assert expected_balance == balance


def get_jetton_address_generate_cassette_name(address, jetton_master_address):
    return (
        f"ton/get_jetton_wallet_address_{address[:8]}_{jetton_master_address[:8]}.yaml"
    )


@pytest.mark.parametrize(
    "address, jetton_master_address, expected_jetton_address, expected_exception",
    [
        (
            "0QDlTHD4T79EyT96gkYNKd3iuRd2__6gGh2PCKpU57jSWbVp",
            "kQAiboDEv_qRrcEdrYdwbVLNOXBHwShFbtKGbQVJ2OKxY_Di",
            "EQAkr7ojLnvr1I894iS4hed-9j0BHr10AgpzI_WMdf8Rnepj",
            None,
        ),
        (
            "0QAEhA1CupMp7uMOUfHHoh7sqAMNu1xQOydf8fQf-ATpkbpT",
            "kQAiboDEv_qRrcEdrYdwbVLNOXBHwShFbtKGbQVJ2OKxY_Di",
            "EQBwxMdpfbD1UuvQeOgeCHUksnuUEeKZo9VxOGmz24Bacw7c",
            None,
        ),
        (
            "EQCYRLyN3G4jePSCOnVVuutLk2pdTCjjSSkGtgOcgZ4GZjHb",
            "kQCKt2WPGX-fh0cIAz38Ljd_OKQjoZE_cqk7QrYGsNP6wfP0",
            "EQAy78ko9RJSTUD0Jrcdb__KlPkcxaPuaAcCaQaXymjyqcDk",
            None,
        ),
        (
            "0QDBorbUtHys99DsbZ4rhfhvE7ddrC1LqUbTmRGNGVEgvVFS",
            "kQCKt2WPGX-fh0cIAz38Ljd_OKQjoZE_cqk7QrYGsNP6wfP0",
            "EQCihRIajzYSzuyWpbROGi5pnv2lvMFcIR_cIgLIuYFWabip",
            None,
        ),
        (
            "0QDGi5-PodzUMc-WzvN7Fv1ysVUpcrAX1xKW9Feb1Y2VD4gl",
            "kQCKt2WPGX-fh0cIAz38Ljd_OKQjoZE_cqk7QrYGsNP6wfP0",
            "EQCh-LEZtAIxMCFGFrIgR0hZSAT37itKz02BHyv92KUCFMHV",
            None,
        ),
        (
            "0QDGi5-PodzUMc-WzvN7Fv1ysVUpc1xKW9Feb1Y2VD4gl",
            "kQCKt2WPGX-fh0cIAz38Ljd_OKQjoZE_cqk7QrYGsNP6wfP0",
            "0",
            InvalidAddressError,
        ),
        (
            "0QDGi5-PodzUMc-WzvN7Fv1ysVUpcrAX1xKW9Feb1Y2VD4gl",
            "QCKt2WPGX-fh0cIAz38Ljd_OKQjoZE_cqk7QrYGsNP6wfP0",
            "0",
            RpcConnectionError,
        ),
    ],
)
async def test_get_jetton_wallet_address(
    ton_client: AioTxTONClient,
    address,
    jetton_master_address,
    expected_jetton_address,
    expected_exception,
):
    cassette_name = get_jetton_address_generate_cassette_name(
        address, jetton_master_address
    )
    with vcr_c.use_cassette(cassette_name):
        if expected_exception:
            with pytest.raises(expected_exception):
                await ton_client.get_jetton_wallet_address(
                    address, jetton_master_address
                )
        else:
            jetton_wallet_address = await ton_client.get_jetton_wallet_address(
                address, jetton_master_address
            )
            assert expected_jetton_address == jetton_wallet_address


@pytest.mark.parametrize(
    "address, limit, lt, hash, to_lt, archival, expected_exception, tx_data_result",
    [
        (
            "0QAEhA1CupMp7uMOUfHHoh7sqAMNu1xQOydf8fQf-ATpkbpT",
            None,
            None,
            None,
            None,
            None,
            None,
            {
                "1Uq/leJuBUpMIqvJiziInWLwNKwG6Si4laiK6RBpUKE=",
                "htFdXopqxYctcjLwfOhXMeRRIaAaoSZ9aRnexbfOKg8=",
                "jtaySLcgIPw1siVV7j4h9YFsO+3nex8cW+3wrJ6dGvw=",
                "T168t1RP6sAi8jXYqqir13ErIcfkIPOIyLJ2K3YGX9A=",
            },
        ),
        (
            "0QAEhA1CupMp7uMOUfHHoh7sqAMNu1xQOydf8fQf-ATpkbpT",
            1,
            None,
            None,
            None,
            None,
            None,
            {"htFdXopqxYctcjLwfOhXMeRRIaAaoSZ9aRnexbfOKg8="},
        ),
        (
            "0QAEhA1CupMp7uMOUfHHoh7sqAMNu1xQOydf8fQf-ATpkbpT",
            1,
            22014131000001,
            "htFdXopqxYctcjLwfOhXMeRRIaAaoSZ9aRnexbfOKg8=",
            None,
            None,
            None,
            {"htFdXopqxYctcjLwfOhXMeRRIaAaoSZ9aRnexbfOKg8="},
        ),
        (
            "Ef9fwskZLEuGDfYTRAtvt9k-mEDkaIskkUOsEwPw1wzXk7zR",
            None,
            None,
            None,
            None,
            None,
            InvalidArgumentError,
            {},
        ),
    ],
)
@vcr_c.use_cassette("ton/get_transactions.yaml")
async def test_get_transactions(
    ton_client: AioTxTONClient,
    address,
    limit,
    lt,
    hash,
    to_lt,
    archival,
    expected_exception,
    tx_data_result,
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await ton_client.get_transactions(address, limit, lt, hash, to_lt, archival)
    else:
        tx_data = await ton_client.get_transactions(
            address, limit, lt, hash, to_lt, archival
        )
        tx_ids = set([shard["transaction_id"]["hash"] for shard in tx_data])
        assert tx_ids.symmetric_difference(tx_data_result) == set()


@pytest.mark.parametrize(
    "address, expected_exception, expected_segno",
    [
        (
            "0QAEhA1CupMp7uMOUfHHoh7sqAMNu1xQOydf8fQf-ATpkbpT",
            None,
            2,
        ),
        (
            "EQBsBYu0o3WC4n3ETH6VyuYcjvQPusd4OhG8VIQFELGzgD1y",
            None,
            1,
        ),
        (
            "EQB1mXHwCmSUfwQDLmdsZ7YvY41i8d8clshV52GjL_v0chUj",
            None,
            0,
        ),
        (
            "Ef9fwskZLEuGDfYTRAtvt9k-mEDkaIskkUOsEwPw1wzXk7zR",
            InvalidArgumentError,
            0,
        ),
        (
            "0:6c058bb4a37582e27dc44c7e95cae61c8ef40fbac7783a11bc54840510b1b380",
            None,
            1,
        ),
    ],
)
@vcr_c.use_cassette("ton/get_transaction_count.yaml")
async def test_get_transaction_count(
    ton_client: AioTxTONClient, address, expected_exception, expected_segno
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await ton_client.get_transaction_count(address)
    else:
        transaction_count = await ton_client.get_transaction_count(address)
        assert transaction_count == expected_segno


@pytest.mark.parametrize(
    "address, expected_exception, expected_packed_address",
    [
        (
            "0:95cc5a9ec1eebbbe1265a20dc63cf015d1143f5871b0ae21eeecd0d60cafc74d",
            None,
            "EQCVzFqewe67vhJlog3GPPAV0RQ/WHGwriHu7NDWDK/HTec6",
        ),
        (
            "0:2e80e2e306b9aefaef7e0a00092a31fd6c8d1abc28d9bf4c5a477e65ef817e30",
            None,
            "EQAugOLjBrmu+u9+CgAJKjH9bI0avCjZv0xaR35l74F+MH/3",
        ),
        (
            "0:6c58bb4a37582e27dc44c7e95cae61c8ef40fbac7783a11bc54840510b1b380",
            None,
            "EQAGxYu0o3WC4n3ETH6VyuYcjvQPusd4OhG8VIQFELGzgG2y",
        ),
        (
            "0:6c58bb4a37582e1bc54840510b1b380",
            None,
            "EQAAAAAAAAAAAAAAAAAAAAAABsWLtKN1guG8VIQFELGzgOMO",
        ),
        (
            "0QAEhA1CupMp7uMOUfHHoh7sqAMNu1xQOydf8fQf+ATpkbpT",
            None,
            "0QAEhA1CupMp7uMOUfHHoh7sqAMNu1xQOydf8fQf+ATpkbpT",
        ),
    ],
)
@vcr_c.use_cassette("ton/pack_address.yaml")
async def test_pack_address(
    ton_client: AioTxTONClient, address, expected_exception, expected_packed_address
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await ton_client.pack_address(address)
    else:
        packed_address = await ton_client.pack_address(address)
        assert packed_address == expected_packed_address


@pytest.mark.parametrize(
    "mnemonics_str, expected_address, expected_raw_address, expected_exception",
    [
        (
            "current worth mimic divert pigeon minor scale abstract bicycle usage talent basic zoo merit melt provide old burger lazy actual amazing drop deposit wink",
            "UQC1-T84mX2mz0gWO-o5ND_mL_aPI1CvgUPQBeXFgVCURVBO",
            "0:b5f93f38997da6cf48163bea39343fe62ff68f2350af8143d005e5c581509445",
            None,
        ),
        (
            "motion churn become nest fault bag clog double please soap damage hen steak nerve letter captain purpose flight aerobic fossil butter asthma hole humble",
            "UQDTd43t-plROumWFwPIeSinJwwaPSatt-ZKbIJzaqlge2tD",
            "0:d3778dedfa99513ae9961703c87928a7270c1a3d26adb7e64a6c82736aa9607b",
            None,
        ),
        (
            "collect maze rough ahead viable upgrade resemble music predict flag movie vocal razor multiply cactus describe host admit battle doctor soldier hungry defy decorate",
            "UQCLeN7G4DNOOTw09ODzLmyfEawMhM1lY5PDga0wbAWBI44G",
            "0:8b78dec6e0334e393c34f4e0f32e6c9f11ac0c84cd656393c381ad306c058123",
            None,
        ),
        (
            "elevator shoulder movie quick rural crime portion pumpkin cattle twenty sound force split example cabbage pen moment curious bitter help attend ocean connect",
            "None",
            "EQAAAAAAAAAAAAAAAAAAAAAABsWLtKN1guG8VIQFELGzgOMO",
            WrongPrivateKey,
        ),
        (
            ["elevator", "wot"],
            "None",
            "0QAEhA1CupMp7uMOUfHHoh7sqAMNu1xQOydf8fQf+ATpkbpT",
            AssertionError,
        ),
    ],
)
@vcr_c.use_cassette("ton/get_address_from_mnemonics.yaml")
async def test_get_address_from_mnemonics(
    ton_client: AioTxTONClient,
    mnemonics_str,
    expected_address,
    expected_raw_address,
    expected_exception,
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await ton_client.get_address_from_mnemonics(mnemonics_str)
    else:
        urlsafe_address, raw_address = await ton_client.get_address_from_mnemonics(
            mnemonics_str
        )
        assert urlsafe_address == expected_address
        assert raw_address == expected_raw_address


@pytest.mark.xfail(
    raises=CannotOverwriteExistingCassetteException,
    reason="boc is always new, we can't test it by VCR",
)
@vcr_c.use_cassette("ton/send_ton_with_custom_seqno.yaml")
async def test_send_ton(ton_client: AioTxTONClient):
    _, address = await ton_client.get_address_from_mnemonics(TON_TEST_WALLET_MEMO)
    tx_count = await ton_client.get_transaction_count(address)
    amount_in_nano = ton_client.to_nano(0.00001)
    result_tx_id = await ton_client.send(
        TON_TEST_WALLET_MEMO,
        "UQDU1hdX6SeHmrvzvyjIrLEWUAdJUJar2sw8haIuT_5n-FLh",
        amount_in_nano,
        seqno=tx_count,
    )
    assert result_tx_id == "GxXCUERZwia9JbvEyb5aqHoaTXdkbk/5R+Kel8tqsjw="


@pytest.mark.xfail(
    raises=CannotOverwriteExistingCassetteException,
    reason="boc is always new, we can't test it by VCR",
)
@vcr_c.use_cassette("ton/send_ton_with_auto_seqno.yaml")
async def test_send_ton_with_auto_seqno(ton_client: AioTxTONClient):
    amount_in_nano = ton_client.to_nano(0.00001)
    recipients_list = [
        {
            "address": "0QAEhA1CupMp7uMOUfHHoh7sqAMNu1xQOydf8fQf-ATpkbpT",
            "amount": amount_in_nano,
            "payload": "Hello, recipient 1!",
            "send_mode": 3,
        },
        {
            "address": "UQDlTHD4T79EyT96gkYNKd3iuRd2__6gGh2PCKpU57jSWQ7j",
            "amount": amount_in_nano,
            "payload": "Hello, recipient 5!",
            "send_mode": 3,
        },
    ]

    with pytest.raises(AssertionError):
        await ton_client.send_bulk(TON_HV_TEST_WALLET_MEMO, recipients_list)

    result_tx_id = await ton_client.send(
        TON_TEST_WALLET_MEMO,
        "UQDU1hdX6SeHmrvzvyjIrLEWUAdJUJar2sw8haIuT_5n-FLh",
        amount_in_nano,
    )
    assert result_tx_id == "aJWE/d/toIz8a3vKut952phyEySylxrKFlqg49qDlQA="


@pytest.mark.xfail(
    raises=CannotOverwriteExistingCassetteException,
    reason="boc is always new, we can't test it by VCR",
)
@vcr_c.use_cassette("ton/send_bulk_ton.yaml")
async def test_send_ton_bulk_seqno(ton_testnet_client_with_hv_wallet: AioTxTONClient):
    amount_in_nano = ton_testnet_client_with_hv_wallet.to_nano(0.0001)
    with pytest.raises(AssertionError):
        await ton_testnet_client_with_hv_wallet.send(
            TON_TEST_WALLET_MEMO,
            "UQDU1hdX6SeHmrvzvyjIrLEWUAdJUJar2sw8haIuT_5n-FLh",
            amount_in_nano,
        )
    recipients_list = [
        {
            "address": "0QAEhA1CupMp7uMOUfHHoh7sqAMNu1xQOydf8fQf-ATpkbpT",
            "amount": amount_in_nano,
            "payload": "Hello, recipient 1!",
            "send_mode": 3,
        },
        {
            "address": "UQDlTHD4T79EyT96gkYNKd3iuRd2__6gGh2PCKpU57jSWQ7j",
            "amount": amount_in_nano,
            "payload": "Hello, recipient 5!",
            "send_mode": 3,
        },
    ]
    result_tx_id = await ton_testnet_client_with_hv_wallet.send_bulk(
        TON_HV_TEST_WALLET_MEMO, recipients_list
    )
    assert result_tx_id == "fzEng6cTXSwD76c75SRMtz6+8/Fy9EGeDBSMKWknjek="


@pytest.mark.xfail(
    raises=CannotOverwriteExistingCassetteException,
    reason="boc is always new, we can't test it by VCR",
)
@vcr_c.use_cassette("ton/deploy_wallet.yaml")
async def test_deploy_wallet(ton_testnet_client_with_hv_wallet: AioTxTONClient):
    result_tx_id = await ton_testnet_client_with_hv_wallet.deploy_wallet(
        TON_HV_TEST_WALLET_MEMO
    )
    assert result_tx_id == "1JYGAdfHBI5NjOl++BIsE9LiMmwS2RPyx2veawNGutg="



import time

from ...boc import Cell
from ._wallet_contract import WalletContract


class WalletV2ContractBase(WalletContract):
    def create_signing_message(self, seqno=None):
        seqno = seqno or 0
        message = Cell()
        message.bits.write_uint(seqno, 32)
        if seqno == 0:
            for _ in range(32):
                message.bits.write_bit(1)
        else:
            timestamp = int(time.time())  # get timestamp in seconds
            message.bits.write_uint(timestamp + 60, 32)

        return message


class WalletV2ContractR1(WalletV2ContractBase):
    def __init__(self, **kwargs) -> None:
        self.code = "B5EE9C724101010100570000AAFF0020DD2082014C97BA9730ED44D0D70B1FE0A4F2608308D71820D31FD31F01F823BBF263ED44D0D31FD3FFD15131BAF2A103F901541042F910F2A2F800029320D74A96D307D402FB00E8D1A4C8CB1FCBFFC9ED54A1370BB6"
        kwargs["code"] = Cell.one_from_boc(self.code)
        super().__init__(**kwargs)


class WalletV2ContractR2(WalletV2ContractBase):
    def __init__(self, **kwargs) -> None:
        self.code = "B5EE9C724101010100630000C2FF0020DD2082014C97BA218201339CBAB19C71B0ED44D0D31FD70BFFE304E0A4F2608308D71820D31FD31F01F823BBF263ED44D0D31FD3FFD15131BAF2A103F901541042F910F2A2F800029320D74A96D307D402FB00E8D1A4C8CB1FCBFFC9ED54044CD7A1"
        kwargs["code"] = Cell.one_from_boc(self.code)
        super().__init__(**kwargs)



from ._cell import Cell
from .dict import serialize_dict


class DictBuilder:
    def __init__(self, key_size: int):
        self.key_size = key_size
        self.items = {}
        self.ended = False

    def store_cell(self, index, value: Cell):
        assert self.ended is False, "Already ended"
        if isinstance(index, bytes):
            index = int(index.hex(), 16)

        assert isinstance(index, int), "Invalid index type"
        assert index not in self.items, f"Item {index} already exist"
        self.items[index] = value
        return self

    def store_ref(self, index, value: Cell):
        assert self.ended is False, "Already ended"

        cell = Cell()
        cell.refs.append(value)
        self.store_cell(index, cell)
        return self

    def end_dict(self) -> Cell:
        assert self.ended is False, "Already ended"
        self.ended = True
        if not self.items:
            return Cell()  # ?

        def default_serializer(src, dest):
            dest.write_cell(src)

        return serialize_dict(self.items, self.key_size, default_serializer)

    def end_cell(self) -> Cell:
        assert self.ended is False, "Already ended"
        assert self.items, "Dict is empty"
        return self.end_dict()


def begin_dict(key_size):
    return DictBuilder(key_size)



import asyncio

from aiotx.clients import AioTxTRONClient

tron_client = AioTxTRONClient(node_url="https://api.shasta.trongrid.io")


async def main():
    wallet = tron_client.generate_address()
    print(wallet)

    balance = await tron_client.get_balance("TEZQQ5BXq3nFKUFJknoV15CW24twzH81La")
    print("balance", balance)

    tx_data = await tron_client.send(
        "private_key",
        "TYge3Gid6vVaQvnPVRJ6SVwzC64cw2eBkN",
        100000,
        memo="Test for Melanty",
    )
    print("tx_data", tx_data)

    tx_data = await tron_client.send_token(
        "private_key",
        "TYge3Gid6vVaQvnPVRJ6SVwzC64cw2eBkN",
        "TG3XXyExBkPp9nzdajDZsozEu4BkaSJozs",
        1999,
        memo="privetpop",
    )
    print("tx_data", tx_data)


if __name__ == "__main__":
    asyncio.run(main())



import asyncio

from aiotx.clients import AioTxLTCClient

ltc_client = AioTxLTCClient(
    node_url="https://api.tatum.io/v3/blockchain/node/litecoin-core-testnet",
    testnet=True,
)


@ltc_client.monitor.on_block
async def handle_block(block):
    print("ltc_client: block", block)


@ltc_client.monitor.on_transaction
async def handle_transaction(transaction):
    print("ltc_client: transaction", transaction)


async def main():
    await ltc_client.start_monitoring()
    while True:
        await asyncio.sleep(1)


if __name__ == "__main__":
    asyncio.run(main())



from ._bit_string import BitString
from ._cell import Cell


class Builder:
    def __init__(self):
        self.bits = BitString(1023)
        self.refs = []
        self.is_exotic = False

    def __repr__(self):
        return "<Builder refs_num: %d, %s>" % (len(self.refs), repr(self.bits))

    def store_cell(self, src: Cell):
        self.bits.write_bit_string(src.bits)
        self.refs += src.refs
        return self

    def store_ref(self, src: Cell):
        self.refs.append(src)
        return self

    def store_maybe_ref(self, src):
        if src:
            self.bits.write_bit(1)
            self.store_ref(src)
        else:
            self.bits.write_bit(0)

        return self

    def store_bit(self, value):
        self.bits.write_bit(value)
        return self

    def store_bit_array(self, value):
        self.bits.write_bit_array(value)
        return self

    def store_uint(self, value, bit_length):
        self.bits.write_uint(value, bit_length)
        return self

    def store_uint8(self, value):
        self.bits.write_uint8(value)
        return self

    def store_int(self, value, bit_length):
        self.bits.write_int(value, bit_length)
        return self

    def store_string(self, value):
        self.bits.write_string(value)
        return self

    def store_bytes(self, value):
        self.bits.write_bytes(value)
        return self

    def store_bit_string(self, value):
        self.bits.write_bit_string(value)
        return self

    def store_address(self, value):
        self.bits.write_address(value)
        return self

    def store_grams(self, value):
        self.bits.write_grams(value)
        return self

    def store_coins(self, value):
        self.bits.write_coins(value)
        return self

    def end_cell(self):
        cell = Cell()
        cell.write_cell(self)
        return cell


def begin_cell():
    return Builder()



import os

import pytest
from conftest import vcr_c

from aiotx.clients import AioTxETHClient
from aiotx.exceptions import (
    AioTxError,
    InvalidArgumentError,
    ReplacementTransactionUnderpriced,
    TransactionNotFound,
    WrongPrivateKey,
)

PRIVATE_KEY_TO_SEND_FROM = os.environ.get("TEST_ETH_WALLET_PRIVATE_KEY")
assert (
    PRIVATE_KEY_TO_SEND_FROM is not None
), "Please provide TEST_ETH_WALLET_PRIVATE_KEY"
CONTRACT = "0x419Fe9f14Ff3aA22e46ff1d03a73EdF3b70A62ED"
DESTINATION_ADDRESS = "0xD515F5737044238B0f0d6c100109DA44066bE749"


@vcr_c.use_cassette("eth/get_last_block.yaml")
async def test_get_last_block(eth_client: AioTxETHClient):
    block_id = await eth_client.get_last_block_number()
    assert isinstance(block_id, int)


@vcr_c.use_cassette("eth/get_chain_id.yaml")
async def test_get_chain_id(eth_client: AioTxETHClient):
    chain_id = await eth_client.get_chain_id()
    assert isinstance(chain_id, int)
    assert chain_id == 11155111


@pytest.mark.parametrize(
    "wallet_address, expected_exception, expected_balance",
    [
        ("0x56ebc43d764761bc09d5918787672e2c8d46da5f", None, 22243749149992258),
        ("0xf663792Be0EdD00AFFB8BBe4Ac6d8185efD5671d", None, 72596311066831845012),
        ("0xA869aF2883DD6C881bC59594aDfB3e8b136321b4", None, 2360954490480603880),
        ("0xE27017DD9EF7e3D0A8321C2041Fc4402e89945b6", None, 9148320984250803464),
        ("0x39832298befa06a6db0c920ada31cfebAD2e33aB", None, 36026722467628523),
        ("0xC4BfcCb1668d6e464f33a76bAdD8c8d7d341E04B", None, 0),
        ("0x68EfbC84d1Eabc193979beab2E2DDc20B219A14", InvalidArgumentError, 0),
    ],
)
@vcr_c.use_cassette("eth/get_balance.yaml")
async def test_get_balance(
    eth_client: AioTxETHClient, wallet_address, expected_exception, expected_balance
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await eth_client.get_balance(wallet_address)
    else:
        balance = await eth_client.get_balance(wallet_address)
        assert isinstance(balance, int)
        assert balance == expected_balance


@pytest.mark.parametrize(
    "tx_id, expected_exception",
    [
        ("0x293fbb7d14e6a11a152fad9d621abdf810acfccb7c8c95977622223638fad96a", None),
        ("0x0d03e84a4d1673b590885ec270553fbd310620debd12d1779796e9ff181f66fe", None),
        ("0x82e8a1058cdadcd0cd728388c48456db7e79051a198d654de85a7f51bf326fb8", None),
        ("0x2f88aa536f9ef65087ce0e81918febf65396cf4a8ff5d535d58a45de4a13d076", None),
        ("0xcb5ad4fd610b6e9e687b5154f4832c0233a527719dcfad6d47500c9f3a50f7cc", None),
        ("0x93ff17f18511c4fa74166a2dbd53ffa7ce5cc2ee431cbb4fd5e02bbdf2701b72", None),
        (
            "0x93ff17f18511c4fa74166a2dbd53ffa7ce5cc2ee431cbb4fd5e02bbdf2701b73",
            TransactionNotFound,
        ),
        (
            "0x93ff17f18511c4fa74166a2dbd53ffa7ce5cc2ee431cbb4fd5e02bbdf2701b2",
            InvalidArgumentError,
        ),
        (
            "0x93ff17f18511c4fa74166a2dbd53ffa7ce5cc2ee431cbb4fd5e02bbdf701b72",
            InvalidArgumentError,
        ),
    ],
)
@vcr_c.use_cassette("eth/get_transaction.yaml")
async def test_get_transaction(eth_client: AioTxETHClient, tx_id, expected_exception):
    if expected_exception:
        with pytest.raises(expected_exception):
            await eth_client.get_transaction(tx_id)
    else:
        tx = await eth_client.get_transaction(tx_id)
        assert "aiotx_decoded_input" in tx.keys()


@pytest.mark.parametrize(
    "wallet_address, contract, expected_exception, expected_balance",
    [
        ("0x159bD467fafcfA348D2b5f357b23aA1d70E814A0", CONTRACT, None, 10109029605),
        ("0x265253254239F1ce54B2580d6Bfa39Afe41D545A", CONTRACT, None, 1665000000),
        ("0xAf0Eca5c36b0380f27412155f6e59618a38e778F", CONTRACT, None, 14000000000),
        ("0x85213732c4F14EaB4B00eB1960a29077302CF131", CONTRACT, None, 12500000000),
        ("0x9cd384d1Dc5Ba08234aE8D57b0d43724caE8C4d6", CONTRACT, None, 11200000000),
        ("0x42827369CD85aD4E88d93D482C144803Aca448C1", CONTRACT, None, 11000000000),
        ("0xDA31b5ac94C559478Eeaa7E347ee4557f9a6F71C", CONTRACT, None, 0),
        (
            "0xDA31b5ac94C559478Eeaa7E347ee4557f9a6F71C",
            "Eeaa7E347ee4557f9a6F71C",
            InvalidArgumentError,
            0,
        ),
    ],
)
@vcr_c.use_cassette("eth/get_token_balance.yaml")
async def test_get_token_balance(
    eth_client: AioTxETHClient,
    wallet_address,
    contract,
    expected_exception,
    expected_balance,
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await eth_client.get_contract_balance(wallet_address, contract)
    else:
        balance = await eth_client.get_contract_balance(wallet_address, contract)
        assert isinstance(balance, int)
        assert expected_balance == balance


@pytest.mark.parametrize(
    "wallet_address, expected_exception, expected_count",
    [
        ("0x159bD467fafcfA348D2b5f357b23aA1d70E814A0", None, 1),
        ("0x265253254239F1ce54B2580d6Bfa39Afe41D545A", None, 2317),
        ("0xAf0Eca5c36b0380f27412155f6e59618a38e778F", None, 358),
        ("0x85213732c4F14EaB4B00eB1960a29077302CF131", None, 809),
        ("0x9cd384d1Dc5Ba08234aE8D57b0d43724caE8C4d6", None, 902),
        ("0xDA31b5ac94C559478Eeaa7E347ee4557f9a6F71C", None, 0),
        ("0xDA31b5ac94C559478Eeaa7E347ee4557f9a6F71", InvalidArgumentError, 0),
        ("DA31b5ac94C559478Eeaa7E347ee4557f9a6F71C", InvalidArgumentError, 0),
    ],
)
@vcr_c.use_cassette("eth/get_transaction_count.yaml")
async def test_get_transaction_count(
    eth_client: AioTxETHClient, wallet_address, expected_exception, expected_count
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await eth_client.get_transactions_count(wallet_address)
    else:
        count = await eth_client.get_transactions_count(wallet_address)
        assert isinstance(count, int)
        assert expected_count == count


@vcr_c.use_cassette("eth/get_gas_price.yaml")
async def test_get_gas_price(eth_client: AioTxETHClient):
    result = await eth_client.get_gas_price()
    assert isinstance(result, int)


@pytest.mark.parametrize(
    "private_key, to_address, amount, gas_price, gas_limit, expected_exception",
    [
        (
            "87e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938chd",
            DESTINATION_ADDRESS,
            0.00001,
            5,
            21000,
            WrongPrivateKey,
        ),
        (
            "87e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938ch",
            DESTINATION_ADDRESS,
            0.00001,
            5,
            21000,
            WrongPrivateKey,
        ),
        (
            "e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938chd",
            DESTINATION_ADDRESS,
            0.00001,
            5,
            21000,
            WrongPrivateKey,
        ),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            "0xf9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b5",
            0.00001,
            5,
            21000,
            ValueError,
        ),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            "f9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b5a",
            0.00001,
            5,
            21000,
            None,
        ),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            "0xf9E35E4e1CbcF08E84d3f6FF662Ba4c306b5a",
            0.00001,
            5,
            21000,
            ValueError,
        ),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            DESTINATION_ADDRESS,
            0.00001,
            5,
            21000,
            ReplacementTransactionUnderpriced,
        ),
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, 5, 5, 21000, AioTxError),
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, 0.00001, 0, 21000, AioTxError),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            DESTINATION_ADDRESS,
            0,
            5,
            21000,
            ReplacementTransactionUnderpriced,
        ),
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, 0.00001, 5, 0, AioTxError),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            DESTINATION_ADDRESS,
            0.00001,
            5,
            61000,
            ReplacementTransactionUnderpriced,
        ),
    ],
)
@vcr_c.use_cassette("eth/send_transaction.yaml")
async def test_send_transaction(
    eth_client: AioTxETHClient,
    private_key,
    to_address,
    amount,
    gas_price,
    gas_limit,
    expected_exception,
):
    """
    Here it's raising ReplacementTransactionUnderpriced and NonceTooLowError because we have reusing
    the same VCR data for every get nonce request, we should investigate how we can change that maybe?
    """
    gas_price = eth_client.to_wei(gas_price, "gwei")
    wei_amount = eth_client.to_wei(amount, "ether")
    if expected_exception:
        with pytest.raises(expected_exception):
            await eth_client.send(
                private_key,
                to_address,
                wei_amount,
                gas_price=gas_price,
                gas_limit=gas_limit,
            )
    else:
        result = await eth_client.send(
            private_key,
            to_address,
            wei_amount,
            gas_price=gas_price,
            gas_limit=gas_limit,
        )
        assert isinstance(result, str)


@pytest.mark.parametrize(
    "private_key, to_address, amount, expected_exception",
    [
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, 0.00001, None),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            "0xf9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b5a",
            0.00001,
            ReplacementTransactionUnderpriced,
        ),
    ],
)
@vcr_c.use_cassette("eth/send_transaction_with_auto_gas.yaml")
async def test_send_transaction_with_auto_gas(
    eth_client: AioTxETHClient, private_key, to_address, amount, expected_exception
):
    """
    Here it's raising ReplacementTransactionUnderpriced and NonceTooLowError because we have reusing
    the same VCR data for every get nonce request, we should investigate how we can change that maybe?
    """
    wei_amount = eth_client.to_wei(amount, "ether")
    if expected_exception:
        with pytest.raises(expected_exception):
            await eth_client.send(private_key, to_address, wei_amount)
    else:
        result = await eth_client.send(private_key, to_address, wei_amount)
        assert isinstance(result, str)


@vcr_c.use_cassette("eth/send_transaction_with_custom_nonce.yaml")
async def test_send_transaction_with_custom_nonce(eth_client: AioTxETHClient):
    wei_amount = eth_client.to_wei(0.00001, "ether")
    sender_address = eth_client.get_address_from_private_key(PRIVATE_KEY_TO_SEND_FROM)
    nonce = await eth_client.get_transactions_count(sender_address)
    first_tx = await eth_client.send(
        PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, wei_amount, nonce=nonce
    )
    assert isinstance(first_tx, str)
    second_tx = await eth_client.send(
        PRIVATE_KEY_TO_SEND_FROM, sender_address, wei_amount, nonce=nonce + 1
    )
    assert isinstance(second_tx, str)


@pytest.mark.parametrize(
    "contract, expected_decimals, expected_exception",
    [
        (CONTRACT, 6, None),
        ("0x337610d27c682E347C9cD60BD4b3b107C9d34dD", 18, InvalidArgumentError),
        ("0x1a0cdabee2c57c965b8bbc037671e458805dfdd5", 18, None),
    ],
)
@vcr_c.use_cassette("eth/get_contract_decimals.yaml")
async def test_get_contract_decimals(
    eth_client: AioTxETHClient, contract, expected_decimals, expected_exception
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await eth_client.get_contract_decimals(contract)
    else:
        decimals = await eth_client.get_contract_decimals(contract)
        assert decimals == expected_decimals


@pytest.mark.parametrize(
    "private_key, to_address, contract, amount, gas_price, gas_limit, expected_exception",
    [
        (
            "87e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938chd",
            DESTINATION_ADDRESS,
            CONTRACT,
            1,
            5,
            61000,
            WrongPrivateKey,
        ),
        (
            "87e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938ch",
            DESTINATION_ADDRESS,
            CONTRACT,
            1,
            5,
            61000,
            WrongPrivateKey,
        ),
        (
            "e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938chd",
            DESTINATION_ADDRESS,
            CONTRACT,
            1,
            5,
            61000,
            WrongPrivateKey,
        ),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            "0xf9E35E4e1CbcF08E984d3f6FF662Ba4c306b5a",
            CONTRACT,
            1,
            5,
            61000,
            ValueError,
        ),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            "f9E5E4e1CbcF08E99B84d3f6FF662Ba4c306b5a",
            CONTRACT,
            1,
            5,
            61000,
            ValueError,
        ),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            "0xf9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b",
            CONTRACT,
            1,
            5,
            61000,
            ValueError,
        ),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            DESTINATION_ADDRESS,
            "0xf9E35E4e1CbcF08E984d3f6FF662Ba4c306b5a",
            0.00001,
            5,
            61000,
            ValueError,
        ),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            DESTINATION_ADDRESS,
            "f9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b5a",
            1,
            5,
            61000,
            None,
        ),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            DESTINATION_ADDRESS,
            "0xf9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b",
            1,
            5,
            61000,
            ValueError,
        ),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            DESTINATION_ADDRESS,
            CONTRACT,
            1,
            5,
            61000,
            ReplacementTransactionUnderpriced,
        ),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            DESTINATION_ADDRESS,
            CONTRACT,
            1000,
            5,
            61000,
            ReplacementTransactionUnderpriced,
        ),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            DESTINATION_ADDRESS,
            CONTRACT,
            1,
            0,
            61000,
            AioTxError,
        ),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            DESTINATION_ADDRESS,
            CONTRACT,
            0,
            5,
            61000,
            ReplacementTransactionUnderpriced,
        ),
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, CONTRACT, 1, 5, 0, AioTxError),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            DESTINATION_ADDRESS,
            CONTRACT,
            1,
            5,
            61000,
            ReplacementTransactionUnderpriced,
        ),
    ],
)
@vcr_c.use_cassette("eth/send_token_transaction.yaml")
async def test_send_token_transaction(
    eth_client: AioTxETHClient,
    private_key,
    to_address,
    contract,
    amount,
    gas_price,
    gas_limit,
    expected_exception,
):
    """
    Here it's raising ReplacementTransactionUnderpriced and NonceTooLowError because we have reusing
    the same VCR data for every get nonce request, we should investigate how we can change that maybe?
    """
    gas_price = eth_client.to_wei(gas_price, "gwei")
    wei_amount = eth_client.to_wei(amount, "mwei")

    if expected_exception:
        with pytest.raises(expected_exception):
            await eth_client.send_token(
                private_key,
                to_address,
                contract,
                wei_amount,
                gas_price=gas_price,
                gas_limit=gas_limit,
            )
    else:
        result = await eth_client.send_token(
            private_key,
            to_address,
            contract,
            wei_amount,
            gas_price=gas_price,
            gas_limit=gas_limit,
        )
        assert isinstance(result, str)


@vcr_c.use_cassette("eth/send_token_transaction_with_custom_nonce.yaml")
async def test_send_token_transaction_with_custom_nonce(eth_client: AioTxETHClient):
    wei_amount = eth_client.to_wei(0.00001, "mwei")
    sender_address = eth_client.get_address_from_private_key(PRIVATE_KEY_TO_SEND_FROM)
    nonce = await eth_client.get_transactions_count(sender_address)
    first_tx = await eth_client.send_token(
        PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, CONTRACT, wei_amount, nonce=nonce
    )
    assert isinstance(first_tx, str)
    second_tx = await eth_client.send_token(
        PRIVATE_KEY_TO_SEND_FROM, sender_address, CONTRACT, wei_amount, nonce=nonce + 1
    )
    assert isinstance(second_tx, str)


@vcr_c.use_cassette("eth/send_token_transaction_with_auto_params.yaml")
async def test_send_transaction_with_auto_params(eth_client: AioTxETHClient):
    wei_amount = eth_client.to_wei(0.00001, "mwei")
    tx_id = await eth_client.send_token(
        PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, CONTRACT, wei_amount
    )
    assert isinstance(tx_id, str)



from setuptools import find_packages, setup


extras_test = [
    "hypothesis",
    "ruff",
    "pyproj",
    "pytest",
    "pytest-cov",
    "sphinx",
    "pytest-asyncio",
    "tox",
    "build",
    "vcrpy",
]

extras_evm = [
    "eth_keys==0.5.*",
    "eth_utils==4.1.*",
    "eth_account==0.13.*",
]

extras_utxo = [
    "sqlalchemy",
    "aiosqlite",
    "aiomysql",
    "greenlet",
    "cryptography",
    "bitcoinlib==0.6.*",
]


setup(
    name="aiotx",
    keywords=[
        "cryptocurrency",
        "blockchain",
        "bitcoin",
        "ethereum",
        "asyncio",
        "aiohttp",
        "json-rpc",
        "payment",
        "wallet",
        "transaction",
    ],
    use_scm_version=True,
    description="An asynchronous library for interacting with various cryptocurrencies and blockchains",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("tests", "scripts", "examples", "docs")),
    package_data={
        "aiotx": ["py.typed"],
        "aiotx.utils": ["*.json"],
    },
    setup_requires=[
        "setuptools_scm",
    ],
    install_requires=[
        "aiohttp",
        "setuptools",
        "pynacl>=1.4.0",
        "tronpy",
    ],
    extras_require={
        "test": extras_test,
        "utxo": extras_utxo,
        "evm": extras_evm,
    },
    url="https://github.com/Grommash9/aiotx",
    project_urls={
        "Documentation": "https://grommash9.github.io/aiotx/",
        "Source": "https://github.com/Grommash9/aiotx",
    },
    author="Oleksandr Prudnikov",
    author_email="prudnikov21@icloud.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
)



import urllib.parse

from ....boc import Cell

SNAKE_DATA_PREFIX = 0x00
CHUNK_DATA_PREFIX = 0x01
ONCHAIN_CONTENT_PREFIX = 0x00
OFFCHAIN_CONTENT_PREFIX = 0x01


def serialize_uri(uri):
    return urllib.parse.quote(uri, safe="~@#$&()*!+=:;,?/'").encode()


def parse_uri(uri):
    return uri.decode()


def create_offchain_uri_cell(uri):
    cell = Cell()
    cell.bits.write_uint8(OFFCHAIN_CONTENT_PREFIX)
    cell.bits.write_bytes(serialize_uri(uri))
    return cell


def parse_offchain_uri_cell(cell):
    assert cell.bits[0] == OFFCHAIN_CONTENT_PREFIX, "Invalid offchain uri cell"
    length = 0
    c = cell
    while c:
        length += len(c.bits)
        c = c.refs[0] if c.refs else None

    _bytes = b""
    length = 0
    c = cell
    while c:
        _bytes += c.bits
        length += len(c.bits)
        c = c.refs[0] if c.refs else None

    return parse_uri(_bytes[1:])



import asyncio

from aiotx.clients import AioTxTONClient

ton_client = AioTxTONClient(
    "https://testnet.toncenter.com/api/v2",
    headers={
        "X-API-Key": "c87fbfad07e66b778bffbe20a5183e55e72f0db97bf68bf5e50170cb53fb115b"
    },
)


@ton_client.monitor.on_block
async def handle_block(block):
    # Here we will get master block id after checking all the shards
    print("ton_client: block", block)


@ton_client.monitor.on_transaction
async def handle_transaction(transaction):
    # here we will get transactions in format:
    """
    {'@type': 'blocks.shortTxId', 'mode': 135,
    'account': '0:cba3b0db18b250e86f96dec1526ce6505d0a918a8fc842facfbc490ec3a68be6',
    'lt': '22079497000007', 'hash': 'I05x4IRYx2AIfWG9fSa/E3S5XF8RxW48Gx3QWEl81CI='}
    """
    # If you want to get more details - you should use get transaction function like that:
    """
    tx_details = await ton_client.get_transactions(
        "0:cba3b0db18b250e86f96dec1526ce6505d0a918a8fc842facfbc490ec3a68be6",
        1, 22079497000007, "I05x4IRYx2AIfWG9fSa/E3S5XF8RxW48Gx3QWEl81CI=")
    """
    # We are not getting all details for all the transaction because it will consume a lot of api calls
    print("ton_client: transaction", transaction)


async def main():
    await ton_client.start_monitoring()
    while True:
        await asyncio.sleep(1)


if __name__ == "__main__":
    asyncio.run(main())
